////////////////////////////////////////////////////////////////////////////
// Lpp.hh - Lpp objects

/////////////////////////////////////////////////////////////////////////////
/* Naming conventions: All member variables end with a capital M.  Private
and protected member functions begin with a Capital letter and their public
counterparts begin with a lowercase letter.  Eg: Cons and cons. */

#ifndef LPP_H
#define LPP_H

/////////////////////////////////////////////////////////////////////////////
// Non "Lpp" prefixed identifiers.

#ifndef LPP_True_NODEFINE
#define True Lpp_True
#endif

#ifndef LPP_S_NODEFINE
#define S Lpp_S
#endif

#ifndef LPP_C_NODEFINE
#define C Lpp_C
#endif

#ifndef LPP_pC_NODEFINE
#define pC Lpp_pC
#endif

#ifndef LPP_EM_NODEFINE
#define EM Lpp_EM
#endif

/////////////////////////////////////////////////////////////////////////////
// Lpp type macros.

#ifdef LPP_NO_TYPE_CHECK
#define the(CN, EXP) ((CN*) EXP)
#define theOr0(CN, EXP) ((CN*) EXP)
#else
#define the(CN, EXP) ((CN*) Lpp_typeCheck(type(CN), EXP))
#define theOr0(CN, EXP) ((CN*) Lpp_typeCheckOr0(type(CN), EXP))
#endif

#define type(CN) Lpp_Type_ ## CN
#define Lpp_theType(CN) CN* the ## CN
#define Lpp_theTypeDecl(CN) Lpp_theType(CN)(let)
#define Lpp_theTypeDef(CN) Lpp_theType(CN)(let obj) {return the(CN, obj);}
#define debugType(CN) Lpp_theTypeDecl(CN)
#define declareType(CN) extern let type(CN)
#define Lpp_colFun(CN) Lpp_gc_ ## CN
#define Lpp_colDecl(CN) void Lpp_colFun(CN)(let)
#define defineType(CN) let type(CN); Lpp_theTypeDef(CN); Lpp_colDecl(CN)
#define initType(CN, SUP, DF) type(CN) = Lpp_type(#CN, SUP, DF)
#define Lpp_colGen(CN) void Lpp_colFun(CN)(let obj) {delete the(CN, obj);}
#ifdef LPP_NO_DEBUG
#define genType(CN, SUP) let initType(CN, SUP, Lpp_colFun(CN))
#else
#define genType(CN, SUP) let initType(CN, SUP, Lpp_colFun(CN));\
                         Lpp_theTypeDef(CN)
#endif
#define genTS(CN, SUP) Lpp_colGen(CN); genType(CN, type(SUP))
#define genT(CN) genTS(CN, Lpp)
#define typeIs(EXP, CN) (typeOf(EXP) == type(CN))
#define classLI(CN) declareType(CN); class CN
#define classL(CN) classLI(CN) : public Lpp
#define CA(CN) Lpp(CN)
#define CT(CN) CA(type(CN))
#define CI(CN, SUP) SUP(type(CN))
#define make(CN) (let) new CN

/////////////////////////////////////////////////////////////////////////////
// Lpp predicate macros.

#define Lpp_pC(EXP) (EXP ? 1 : 0)

/////////////////////////////////////////////////////////////////////////////
// Class Lpp is the base class of all Lpp derived classes and super most type.

class Type; class Lpp;

typedef Lpp* let;

classLI(Lpp) {
  let typeM;
protected:
  Lpp(let);
  int Typep(let); const char* TypeName(); let TypeOf(); void SetType(let);
public:
  friend void Lpp_gc_Lpp(let);
  friend let typep(let, let); friend void Lpp_setType(let, let);
  friend const char* typeName(let); friend let typeOf(let);
  friend let Lpp_typeCheck(let, let);
  friend let Lpp_typeCheckOr0(let, let);
  friend let identity(let);};

/////////////////////////////////////////////////////////////////////////////
// Prototypes in advance (just for sequencing of code).

extern let Lpp_True;
let funcall(let, let, let);
typedef void (*Lpp_vfp1)(let);

/////////////////////////////////////////////////////////////////////////////
// Class Type is the dynamic type component of Lpp classes.

classL(Type) {
  const char* nameM; let supersM;
  let princFunctionM; let prin1FunctionM;
  let equalFunctionM; Lpp_vfp1 gcFunctionM;
  let puthashFunctionM; let gethashFunctionM;
protected:
  Type(const char*, let); const char* TypeName();
  let TypeSupers(); int TypeContainsp(let);
  void SetTypePrinc(let); let TypePrinc();
  void SetTypePrin1(let); let TypePrin1();
  void SetTypeEqual(let);
  void SetTypePuthash(let); let TypePuthash();
  void SetTypeGethash(let); let TypeGethash();
public:
  friend let copy(let); friend char* copy(const char*);
  friend void Lpp_gc_Type(let); friend void gc(let);
  friend void gcList(let, let); friend void gcTree(let, let);
  friend void gcListC(let, let); friend void gcTreeC(let, let);
  friend let Lpp_type(const char*, let, Lpp_vfp1);
  friend const char* Lpp_typeName(let);
  friend int Lpp_typeContainsp(let, let);
  friend let setTypePrin1(let, let); friend let typePrin1(let);
  friend let setTypePrinc(let, let); friend let typePrinc(let);
  friend let setTypePrins(let, let);
  friend let Lpp_setTypePuthash(let, let); friend let Lpp_typePuthash(let);
  friend let Lpp_setTypeGethash(let, let); friend let Lpp_typeGethash(let);
  friend void setTypeEqual(let, let);
  friend let eq(let x, let y) {return x == y ? Lpp_True : 0;}
  friend let eql(let, let); friend let equal(let, let);};

debugType(Type);

/////////////////////////////////////////////////////////////////////////////
// Initialization of Lpp type and user types initialized at top level.

class Lpp_initialization {
  static int initialized;
public:
  Lpp_initialization();};

static Lpp_initialization Lpp_initialization_instance;

void topTypes(let, ...);

/////////////////////////////////////////////////////////////////////////////
// Symbol object.

classL(Symbol) {
  let nameM; static let Lpp_nameTableM;
protected:
  Symbol(let); let SymbolName();
public:
  friend void Lpp_gc_Symbol(let); friend let Lpp_uninternedSymbol(let);
  friend void Lpp_initDispatchSymbol(); friend void Lpp_initTypeSymbol();
  friend let symbolName(let s); friend let symbolp(let);
  friend let intern(let); friend let intern(const char*);};

debugType(Symbol);
#define Lpp_S(SN) intern(#SN)

/////////////////////////////////////////////////////////////////////////////
// Hash table object.

classL(HashTable) {
  int sizeM; let* tableM; let testFunM;
protected:
  HashTable(int, let);
public:
  friend void Lpp_gc_HashTable(let);
  friend let hashTableP(let);
  friend let makeHashTable(); friend let makeHashTable(int); friend let makeHashTable(int, let);
  friend let puthash(let, let, let);
  friend let Lpp_puthashString(let, let, let);
  friend let Lpp_puthashSymbol(let, let, let);
  friend let gethash(let, let); friend let gethash(let, let, let);
  friend let Lpp_gethashString(let, let);
  friend let Lpp_gethashSymbol(let, let);
  friend let remhash(let, let); friend let clrhash(let);
  friend unsigned sxhash(let);};

void Lpp_initTypeHashTables();

debugType(HashTable);

/////////////////////////////////////////////////////////////////////////////
// Function object.

#define LPP_LET2 let,let
#define LPP_LET3 let,let,let
#define LPP_LET4 let,let,let,let
#define LPP_LET5 let,let,let,let,let
#define LPP_LET6 let,let,let,let,let,let
#define LPP_LET7 let,let,let,let,let,let,let
#define LPP_LET8 let,let,let,let,let,let,let,let
#define LPP_LET9 let,let,let,let,let,let,let,let,let
#define LPP_LET10 let,let,let,let,let,let,let,let,let,let

typedef let (*Lpp_fp0)();
typedef let (*Lpp_fp1)(let); typedef let (*Lpp_fp2)(LPP_LET2);
typedef let (*Lpp_fp3)(LPP_LET3); typedef let (*Lpp_fp4)(LPP_LET4);
typedef let (*Lpp_fp5)(LPP_LET5); typedef let (*Lpp_fp6)(LPP_LET6);
typedef let (*Lpp_fp7)(LPP_LET7); typedef let (*Lpp_fp8)(LPP_LET8);
typedef let (*Lpp_fp9)(LPP_LET9); typedef let (*Lpp_fp10)(LPP_LET10);

classL(Function) {
protected:
  Function(); int argCountM;
public:
  friend void Lpp_gc_Function(let);
  friend let functionp(let);
  friend let apply(let, let); friend argCount(let);};

debugType(Function);

class Function0 : private Function {let (*functionM)(); Function0(Lpp_fp0);
public: friend let Lpp_C(Lpp_fp0); friend let funcall(let);};
class Function1 : private Function {let (*functionM)(let); Function1(Lpp_fp1);
public: friend let Lpp_C(Lpp_fp1); friend let funcall(let, let);};
class Function2 : private Function {let (*functionM)(LPP_LET2); Function2(Lpp_fp2);
public: friend let Lpp_C(Lpp_fp2); friend let funcall(let, LPP_LET2);};
class Function3 : private Function {let (*functionM)(LPP_LET3); Function3(Lpp_fp3);
public: friend let Lpp_C(Lpp_fp3); friend let funcall(let, LPP_LET3);};
class Function4 : private Function {let (*functionM)(LPP_LET4); Function4(Lpp_fp4);
public: friend let Lpp_C(Lpp_fp4); friend let funcall(let, LPP_LET4);};
class Function5 : private Function {let (*functionM)(LPP_LET5); Function5(Lpp_fp5);
public: friend let Lpp_C(Lpp_fp5); friend let funcall(let, LPP_LET5);};
class Function6 : private Function {let (*functionM)(LPP_LET6); Function6(Lpp_fp6);
public: friend let Lpp_C(Lpp_fp6); friend let funcall(let, LPP_LET6);};
class Function7 : private Function {let (*functionM)(LPP_LET7); Function7(Lpp_fp7);
public: friend let Lpp_C(Lpp_fp7); friend let funcall(let, LPP_LET7);};
class Function8 : private Function {let (*functionM)(LPP_LET8); Function8(Lpp_fp8);
public: friend let Lpp_C(Lpp_fp8); friend let funcall(let, LPP_LET8);};
class Function9 : private Function {let (*functionM)(LPP_LET9); Function9(Lpp_fp9);
public: friend let Lpp_C(Lpp_fp9); friend let funcall(let, LPP_LET9);};
class Function10 : private Function {let (*functionM)(LPP_LET10); Function10(Lpp_fp10);
public: friend let Lpp_C(Lpp_fp10); friend let funcall(let, LPP_LET10);};
   
/////////////////////////////////////////////////////////////////////////////
// Cons cell object.

extern let Lpp_EM; extern let Lpp_CM;

classL(Cons) {
  let carM; let cdrM;
protected:
  Cons(let, let); let Car(); let Cdr(); let Rplaca(let); let Rplacd(let);
public:
  friend void Lpp_gc_Cons(let);
  friend void Lpp_initDispatchCons();
  friend let cons(let, let); friend let car(let); friend let cdr(let);
  friend let rplaca(let, let); friend let rplacd(let, let);
  friend let consEQ(let, let);
  friend let list(); friend let list(let); friend let list(LPP_LET2); friend let list(LPP_LET3); friend let list(LPP_LET4); friend let list(LPP_LET5); friend let list(LPP_LET6); friend let list(LPP_LET7); friend let list(LPP_LET8); friend let list(LPP_LET9); friend let list(LPP_LET10);
  friend let listEM(let, ...); friend let listSEM(const char*, ...);
  friend let consp(let); friend let listp(let);
  friend int listLength(let); friend let length(let);
  friend let last(let); friend let last(let, int);
  friend let copyList(let); friend let copyAlist(let); friend let copyTree(let);
  friend let nth(int, let); friend let nthcdr(int, let);
  friend let first(let list) {return car(list);} friend let second(let); friend let third(let); friend let fourth(let); friend let fifth(let); friend let sixth(let); friend let seventh(let); friend let eighth(let); friend let ninth(let); friend let tenth(let);
  friend let member(let, let); friend let member(let, let, let);
  friend let assoc(let, let); friend let assoc(let, let, let);
  friend let remove(let, let); friend let remove(let, let, let); 
  friend let nremove(let, let, let); friend let nremove(let, let, let, let); 
  friend let nconc(let, let); friend let listConcat(let, let);
  friend let reverse(let); friend let nreverse(let);
  friend let mapcar(let, let); friend let mapcar(let, let, let);
  friend let maplist(let, let); friend let maplist(let, let, let);
  friend let mapc(let, let); friend let mapc(let, let, let);
  friend let mapl(let, let); friend let mapl(let, let, let);
  friend let mapcan(let, let); friend let mapcan(let, let, let);
  friend let mapcon(let, let); friend let mapcon(let, let, let);};

debugType(Cons);

/////////////////////////////////////////////////////////////////////////////
// Cons macros.

#define push(EL, LIST) (LIST = cons(EL, LIST))
#define pop(EL, LIST) EL = car(LIST); LIST = cdr(LIST)
#define Lpp_lv(VAR) Lpp_lv_ ## VAR
#define dolist(VAR, EXP) \
  for(let Lpp_lv(VAR) = EXP, VAR = car(Lpp_lv(VAR)); \
      Lpp_lv(VAR); \
      Lpp_lv(VAR) = cdr(Lpp_lv(VAR)), VAR = car(Lpp_lv(VAR)))
#define dolist2(VAR1, EXP1, VAR2, EXP2) \
  for(let Lpp_lv(VAR1) = EXP1, Lpp_lv(VAR2) = EXP2, \
          VAR1 = car(Lpp_lv(VAR1)), VAR2 = car(Lpp_lv(VAR2)); \
      Lpp_lv(VAR1) && Lpp_lv(VAR2); \
      Lpp_lv(VAR1) = cdr(Lpp_lv(VAR1)), \
      Lpp_lv(VAR2) = cdr(Lpp_lv(VAR2)), \
      VAR1 = car(Lpp_lv(VAR1)), VAR2 = car(Lpp_lv(VAR2)))
#define doCdrs(VAR, EXP) for(let VAR = EXP; VAR; VAR = cdr(VAR))
#define doCdrs2(VAR1, EXP1, VAR2, EXP2) \
  for(let VAR1 = EXP1, VAR2 = EXP2; VAR1 && VAR2; \
        VAR1 = cdr(VAR1), VAR2 = cdr(VAR2))

/////////////////////////////////////////////////////////////////////////////
// String object.

classL(String) {
  int lengthM; char* cstringM;
protected:
  String(const char*); String(const char*, int);
  const char* StringC();
  int Length();
public:
  friend void Lpp_gc_String(let);
  friend void Lpp_initDispatchString();
  friend let Lpp_C(const char *); friend const char* sC(let);
  friend let stringp(let);
  friend int stringLength(let); friend let lengthString(let);
  friend let stringEqual(let, let);
  friend let stringConcat(let, let);};

debugType(String);

/////////////////////////////////////////////////////////////////////////////
// Number object.

classL(Number) {
protected:
  Number(let type) : CA(type) {}
public:
  friend void Lpp_gc_Number(let);
  friend let numberp(let);
  friend let plus(let, let); friend let minus(let, let);
  friend let times(let, let); friend let divide(let, let);
  friend let inc(let); friend let inc(let, let);
  friend let dec(let); friend let dec(let, let);
  friend let lessThan(let, let); friend let greaterThan(let, let);
  friend let lessThanOrEqual(let, let);
  friend let greaterThanOrEqual(let, let);};

/////////////////////////////////////////////////////////////////////////////
// Integer object.

classLI(Integer) : private Number {
  int intM;
  Integer(int);
protected:
  int IntC(); void Inc(let); void Dec(let);
public:
  friend void Lpp_gc_Integer(let);
  friend void Lpp_initDispatchInteger();
  friend let Lpp_C(int); friend int iC(let); friend let integerp(let);
  friend let equalI(let, let);
  friend let plusI(let, let); friend let minusI(let, let);
  friend let timesI(let, let); friend let divideI(let, let);
  friend let incI(let); friend let incI(let, let);
  friend let decI(let); friend let decI(let, let);
  friend let lessThanI(let, let); friend let greaterThanI(let, let);
  friend let lessThanOrEqualI(let, let);
  friend let greaterThanOrEqualI(let, let);};
      
debugType(Integer);

/////////////////////////////////////////////////////////////////////////////
// Input/Output

void Lpp_initIO();
extern let defaultPrin1Address;
let defaultPrin1Function(); let error(let, ...); let error(const char*, ...);
let princ(let); let prin1(let); let print(let); let terpri();
const char* princ(const char*); const char* prin1(const char*);
const char* print(const char*);
let prin1ToString(let); let princToString(let);
int prin1Length(let); void pprint(let);
let readFromString(let); let readFromString(const char*);
let finishOutput(); let pdb(let); let ppdb(let); let pdc(let);
let read();

/////////////////////////////////////////////////////////////////////////////
// Environment Inquiries.

let currentDirectory();

/////////////////////////////////////////////////////////////////////////////
#endif LPP_H
